<?php
/**
 * DokuWiki Plugin authelkarte (Auth Component)
 *
 * Provides authentication against a ElkArte MySQL database backend
 *
 * @name      ElkArte Forum
 * @copyright ElkArte Forum contributors
 * @license   BSD http://opensource.org/licenses/BSD-3-Clause
 *
 * @version 1.0
 *
 */

// Must be run within Dokuwiki
if (!defined('DOKU_INC'))
{
	die();
}

/**
 * ElkArte password hashing functions
 */
require_once('class-phpass.php');

/**
 * Authentication class
 */
class auth_plugin_authelkarte extends DokuWiki_Auth_Plugin
{
	/**
	 * SQL statement to retrieve member data from Elkarte DB
	 *
	 * - '%prefix%' will be replaced by the actual prefix (from plugin config)
	 */
	protected $sql_elk_user_data = "
		SELECT
            id_member, member_name, passwd, email_address, real_name, id_group, id_post_group, additional_groups
        FROM `%prefix%members`";

	protected $sql_elk_group_data = "
		SELECT
            id_group, group_name
        FROM `%prefix%membergroups`";

	/**
	 * ElkArte database connection
	 * @var mysqli
	 */
	protected $db;

	/**
	 * Users cache
	 */
	protected $users;

	/**
	 * If all users have been loaded in the cache
	 */
	protected $usersCached = false;

	/**
	 * Filter pattern
	 */
	protected $filter;

	/**
	 * Group Names
	 */
	protected $group_names;

	/**
	 * Constructor.
	 */
	public function __construct()
	{
		parent::__construct();

		// Define this plugin capabilities, not many for now
		$this->cando['getUsers'] = true;
		$this->cando['getUserCount'] = true;
		$this->cando['getGroups'] = true;

		// @todo use to check if they are already logged into elk after witting the trustExternal function
		// https://www.dokuwiki.org/devel:auth_plugins
		//$this->cando['external'] = true;

		// Establish a connection to the ElkArte DB
		try
		{
			$this->connectElkarteDb();
		}
		catch (Exception $e)
		{
			// Can't connect so abort
			msg(sprintf($this->getLang('error_connect_failed'), $e->getMessage()));
			$this->success = false;

			return;
		}

		// Initialize our SQL querys with the prefix set in config file
		$this->sql_elk_user_data = str_replace(
			'%prefix%',
			$this->getConf('prefix'),
			$this->sql_elk_user_data
		);

		$this->sql_elk_group_data = str_replace(
			'%prefix%',
			$this->getConf('prefix'),
			$this->sql_elk_group_data
		);

		$this->success = true;
	}

	/**
	 * Connect to ElkArte database
	 *
	 * - Initializes $db property as mysqli object
	 */
	protected function connectElkarteDb()
	{
		// Already have a connection?
		if ($this->db)
		{
			return;
		}

		// Create the connection
		if ($this->getConf('port'))
		{
			$this->db = new mysqli($this->getConf('hostname'), $this->getConf('username'), $this->getConf('password'), $this->getConf('database'), $this->getConf('port'));
		}
		else
		{
			$this->db = new mysqli($this->getConf('hostname'), $this->getConf('username'), $this->getConf('password'), $this->getConf('database'));
		}
	}

	/**
	 * Check user has entered thier correct password
	 *
	 * @param string $user the user name
	 * @param string $pass the password
	 *
	 * @return bool
	 *
	 * @uses PasswordHash::CheckPassword
	 */
	public function checkPass($user, $pass)
	{
		// Fetch the user data to check
		$data = $this->getUserData($user);

		if ($data === false)
		{
			return false;
		}

		// Use Elkarte hashing / validation on the password
		$hasher = new PasswordHash(10, false);
		$pass = hash('sha256', strtolower($user) . $pass);
		$check = (bool) $hasher->CheckPassword($pass, $data['pass']);

		// Log and return the results
		dbglog("Password " . ($check ? 'OK' : 'Invalid'));

		return $check;
	}

	/**
	 * Returns basic information about a user
	 *
	 * @param string $user the user name
	 * @param bool $requireGroups
	 *
	 * @return array
	 */
	public function getUserData($user, $requireGroups = true)
	{
		// Cached? then return the data
		if (isset($this->users[$user]))
		{
			return $this->users[$user];
		}

		// Define the query
		$sql = $this->sql_elk_user_data . ' WHERE `member_name` = \'' . $this->db->escape_string($user) . '\'';

		// Run the query
		$result = $this->db->query($sql);

		// Make sure it executed
		if (!$result)
		{
			// Problem making the query, lets log it
			$err = 'Error ' . $this->db->error . ': ' . $this->db->errno;
			dbglog($err);

			return false;
		}

		// Fetch query results
		$user = $result->fetch_assoc();

		// Did we find them?
		if ($user === false)
		{
			// Unknown user
			dbglog("Unknown user");

			return false;
		}

		return $this->cacheUser($user);
	}

	/**
	 * Translate ElkArte user data to DokuWiki user info array
	 *
	 * - Stores result in the users cache
	 *
	 * @param array $row
	 *
	 * @return array user data
	 */
	protected function cacheUser($row)
	{
		global $conf;

		$login = $row['member_name'];

		// If the user is already cached, just return it
		if (isset($this->users[$login]))
		{
			return $this->users[$login];
		}

		// Group membership
		$groups = array_merge(array($row['id_group'], $row['id_post_group']), explode(',', $row['additional_groups']));
		$groups = array_filter($groups);

		// Elkarte group 1 is the admin group ...
		foreach ($groups as $id => $group)
		{
			if ($group == 1)
			{
				$groups[$id] = 'admin';
			}
			else
			{
				$groups[$id] = $this->group_names[$group];
			}
		}

		// Give them default groups if defined.
		if ($this->getConf('usedefaultgroup'))
		{
			$groups[] = $conf['defaultgroup'];
		}

		// Set it up so DokuWiki understands the data
		$info = array(
			'user' => $login,
			'name' => $row['real_name'],
			'pass' => $row['passwd'],
			'mail' => $row['email_address'],
			'grps' => $groups,
		);

		$this->users[$login] = $info;

		return $info;
	}

	/**
	 * Bulk retrieval of user data for display in DokuWiki
	 *
	 * @param int $start index of first user to be returned
	 * @param int $limit max number of users to be returned
	 * @param array $filter array of field/pattern pairs
	 *
	 * @return  array
	 */
	public function retrieveUsers($start = 0, $limit = 0, $filter = array())
	{
		// Let them know this is a read only list
		msg($this->getLang('user_list_use_elkarte'));

		// Prefer not to do this too often
		$this->cacheAllUsers();

		// Apply any filter and pagination
		$this->setFilter($filter);
		$list = array();
		$count = 0;
		$i = 0;

		// For all the users
		foreach ($this->users as $user => $info)
		{
			if ($this->applyFilter($user, $info))
			{
				if ($i >= $start)
				{
					$list[$user] = $info;
					$count++;

					if ($limit > 0 && $count >= $limit)
					{
						break;
					}
				}

				$i++;
			}
		}

		return $list;
	}

	/**
	 * Loads all ElkArte users into the cache
	 */
	protected function cacheAllUsers()
	{
		if ($this->usersCached)
		{
			return;
		}

		// Group names we can understand
		$this->group_names = $this->retrieveGroups();

		// Execute our mysqli statement
		$result = $this->db->query($this->sql_elk_user_data);
		while ($user = $result->fetch_assoc())
		{
			$this->cacheUser($user);
		}

		$this->usersCached = true;
	}

	/**
	 * Fetch the list of ElkArte groups
	 *
	 * @param int $start
	 * @param int $limit
	 *
	 * @return array|bool
	 */
	public function retrieveGroups($start = 0, $limit = 0)
	{
		if ($limit > 0)
		{
			$sql = $this->sql_elk_group_data . " WHERE 1 = 1 ORDER BY 'group_name' LIMIT $start $limit";
		}
		else
		{
			$sql = $this->sql_elk_group_data . " ORDER BY 'group_name'";
		}

		// Get groups from db
		$result = $this->db->query($sql);
		$tmpvar = array();
		if ($result)
		{
			while ($row = $result->fetch_assoc())
			{
				$tmpvar[$row['id_group']] = $row['group_name'];
			}
		}

		return $tmpvar;
	}

	/**
	 * Build filter patterns from given criteria
	 *
	 * @param array $filter
	 */
	protected function setFilter($filter)
	{
		$this->filter = array();
		foreach ($filter as $field => $value)
		{
			// Build regex filter
			$this->filter[$field] = '/' . str_replace('/', '\/', $value) . '/ui';
		}
	}

	/**
	 * Return if given user matches filter pattern
	 *
	 * @param string $user login
	 * @param array $info User data
	 *
	 * @return bool
	 */
	protected function applyFilter($user, $info)
	{
		foreach ($this->filter as $elem => $pattern)
		{
			if ($elem === 'grps')
			{
				$result = preg_grep($pattern, $info['grps']);
				if (empty($result))
				{
					return false;
				}
			}
			else
			{
				if (!preg_match($pattern, $info[$elem]))
				{
					return false;
				}
			}
		}

		return true;
	}

	/**
	 * Return the count users which meet $filter criteria
	 *
	 * @param array $filter
	 *
	 * @return int
	 */
	public function getUserCount($filter = array())
	{
		$this->cacheAllUsers();
		$count = 0;

		if (empty($filter))
		{
			$count = count($this->users);
		}
		else
		{
			$this->setFilter($filter);
			foreach ($this->users as $user => $info)
			{
				$count += (int) $this->applyFilter($user, $info);
			}
		}

		return $count;
	}
}